#include "db_compiler.h"

#define OPERAND_STACK_SIZE  10
#define OPERATOR_STACK_SIZE 10

typedef struct {
    Token operator;
    int precedence;
} Operator;

static ParseTreeNode operandStack[OPERAND_STACK_SIZE];
static int ioperand;

static Operator operatorStack[OPERATOR_STACK_SIZE];
static int ioperator;

static ParseTreeNode *PushOperand(ParseContext *c, int nodeType);
static void PushOperator(ParseContext *c, Token operator, int precedence);
static int GetPrecedence(Token tkn, int *pLeftAssociative);
static int IsUnary(Token tkn);
static void xGetSymbolRef(ParseContext *c, char *name, ParseTreeNode *node);

/* ParseExpr - parse an expression */
ParseTreeNode *ParseExprNR(ParseContext *c)
{
    ParseTreeNode *node;
    int done = VMFALSE;
    int precedence;
    int leftAssociative;
    Token tkn;
    
    /* clear the stacks */
    ioperand = ioperator = -1;
    
    /* parse an expression */
    for (;;) {
        switch (tkn = GetToken(c)) {
        case T_IDENTIFIER:
            node = PushOperand(c, NodeTypeSymbolRef);
            xGetSymbolRef(c, c->token, node);
            break;
        case T_NUMBER:
            node = PushOperand(c, NodeTypeIntegerLit);
            node->integerLit.value = c->value;
            break;
        case T_STRING:
            node = PushOperand(c, NodeTypeStringLit);
            node->stringLit.string = AddString(c, c->token);
            break;
        case '(':
            PushOperator(c, '(', 0);
            break;
        case ')':
            break;
        default:
            if ((precedence = GetPrecedence(tkn, &leftAssociative)) == 0)
                done = VMTRUE;
            else {
                while (ioperator >= 0) {
                    if ((leftAssociative && precedence <= operatorStack[ioperator].precedence)
                    ||  (!leftAssociative && precedence < operatorStack[ioperator].precedence))
                        ;
                }
                PushOperator(c, tkn, precedence);
            }
            break;
        }
    }
}

/* PushOperand - push an entry on the operand stack */
static ParseTreeNode *PushOperand(ParseContext *c, int nodeType)
{
    ParseTreeNode *node;
    if (++ioperand >= OPERAND_STACK_SIZE)
        ParseError(c, "operand stack overflow");
    node = &operandStack[ioperand];
    node->nodeType = nodeType;
    return node;
}

/* PushOperator - push an entry on the operand stack */
static void PushOperator(ParseContext *c, Token operator, int precedence)
{
    if (++ioperator >= OPERAND_STACK_SIZE)
        ParseError(c, "operand stack overflow");
    operatorStack[ioperator].operator = operator;
    operatorStack[ioperator].precedence = precedence;
}

/* GetPrecedence - get the precedence of an operator */
static int GetPrecedence(Token tkn, int *pLeftAssociative)
{
    int precedence = 0;
    *pLeftAssociative = VMTRUE;
    switch (tkn) {
    case T_OR:
        precedence = 1;
        break;
    case T_AND:
        precedence = 2;
        break;
    case '^':
        precedence = 3;
        break;
    case '|':
        precedence = 4;
        break;
    case '&':
        precedence = 5;
        break;
    case '=':
    case T_NE:
        precedence = 6;
        break;
    case '<':
    case T_LE:
    case T_GE:
    case '>':
        precedence = 7;
        break;
    case T_SHL:
    case T_SHR:
        precedence = 8;
        break;
    case '+':
    case '-':
        precedence = 9;
        break;
    case '*':
    case '/':
    case T_MOD:
        precedence = 10;
        break;
    }
    return precedence;
}

/* IsUnary - check for unary operators */
static int IsUnary(Token tkn)
{
    return tkn == '+' || tkn == '-' || tkn == T_NOT || tkn == '~';
}

/* GetSymbolRef - setup a symbol reference */
static void xGetSymbolRef(ParseContext *c, char *name, ParseTreeNode *node)
{
    Symbol *symbol;

    /* handle local variables within a function or subroutine */
    if (c->codeType != CODE_TYPE_MAIN && (symbol = FindSymbol(&c->locals, name)) != NULL) {
        if (IsConstant(symbol)) {
            node->nodeType = NodeTypeIntegerLit;
            node->integerLit.value = symbol->value;
        }
        else {
            node->symbolRef.symbol = symbol;
            node->symbolRef.fcn = code_local;
            node->symbolRef.offset = symbol->value;
        }
    }

    /* handle function or subroutine arguments or the return value symbol */
    else if (c->codeType != CODE_TYPE_MAIN && (symbol = FindSymbol(&c->arguments, name)) != NULL) {
        node->symbolRef.symbol = symbol;
        node->symbolRef.fcn = code_local;
        node->symbolRef.offset = symbol->value - c->arguments.count;
    }

    /* handle global symbols */
    else if ((symbol = FindSymbol(&c->globals, c->token)) != NULL) {
        if (IsConstant(symbol)) {
            node->nodeType = NodeTypeIntegerLit;
            node->integerLit.value = symbol->value;
        }
        else {
            node->symbolRef.symbol = symbol;
            node->symbolRef.fcn = code_global;
            node->symbolRef.offset = symbol->value;
        }
    }

    /* handle undefined symbols */
    else {

        /* if inside a function or subroutine definition, add the symbol as a local */
        if (c->codeType != CODE_TYPE_MAIN) {
            symbol = AddLocal(c, name, SC_VARIABLE, c->localOffset++ + F_SIZE + 1);
            node->symbolRef.symbol = symbol;
            node->symbolRef.fcn = code_local;
            node->symbolRef.offset = symbol->value;
        }

        /* otherwise, add it as a global */
        else {
            symbol = AddGlobal(c, name, SC_VARIABLE, c->image->variableCount++, 0);
            node->symbolRef.symbol = symbol;
            node->symbolRef.fcn = code_global;
            node->symbolRef.offset = symbol->value;
        }
    }
}

